import os
import copy
import StringIO

"""
import logging
logging.basicConfig(level=logging.DEBUG)
"""

from knuth import knuth
from DviFont import *
import VirtualFont

import subprocess

def Kstr(x):
	return("%7.2f" % x)

"""
.dvi file reading:
	a .dvi file is a string of bytes
	partitioned into commands
	each command is an array of bytes
	the initial one is the command index;
		near the beginning the total length of the command 
		and maybe the number of bytes in various parts is specified 

	While reading the file, the state is essentially specified by
		(a) the current font
		(b) an array [ h, v, w, x, y, z ]
		(c) a stack of such arrays
	but in fact I keep track of various other things, too. 

	[h, v] is the current position
	[w, x, y, z] specify certain dimensions

	The coordinate system has origin at upper left, v-axis going down.
		This is because then it doesn't have to know what the page size is.
		The graphics rendering program might do this;
		for example DviEPS does, since its coordinate system is at lower left.
	Variables are of two kinds, although confounded in a .dvi file:
		integers and lengths.  	

		Units of length are sp units, but subject to magnification.  
		There are 2^16 sp in a Knuth point, 72.27 Knuth points to an inch, 
		which is also equals 72 Adobe points.  A length in Adobe points is obtained from that in sp
		by multiplying it by (mag/1000)*2^{-16}*(72/72.27).

		The design size of a font is specified in sp units.
		The width, height, depth of characters is specified
		as a fraction of font design size, scaled by 2^20.
		So a true character dimension in sp units is
		(font design size)*(character width).

	The command index is in the range [0, 256), so equals one byte
	The data following the index is of two basic types:
		(a) integers of 1-4 bytes
		(b) arrays of bytes, usually strings of characters
	Integers of 1-3 bytes are either signed or unsigned.
		Those of 4 are signed.  These can be either lengths or integers.

	Each command reads its data, but it does not know
		how to use it.  The state the commands change is
		managed by the DviReader, so each command just calls
		back to the Reader to execute.  The most basic DviReader
		just keeps track of where we are on a page,
		also current dimensions, and current fonts.
		It uses stacks to do this.  More advanced Readers such as DviEPS or DviJava
		know how to display the page - this depends on
		what kind of display we are dealing with.

		March 23, 2008: I now plant rules.

Some of the command indices are:

"""

SETRULE = 132
PUTRULE =137
NOP = 138
BOP = 139
EOP = 140
PUSH = 141
POP = 142
# W, X, Y, Z, RIGHT, DOWN
FNT1 = 235
FNT2 = 236
FNT3 = 237
FNT4 = 238
XXX1 = 239
XXX2 = 240
XXX3 = 241
XXX4 = 242
FNTDEF1 = 243
FNTDEF2 = 244
FNTDEF3 = 245
FNTDEF4 = 246
PRE = 247
POST = 248
POSTPOST = 249

# =====================================================================================
# === now the different commands ======================================================
# === each command is its own class, subclassing DviCommand ===========================

# in first version: read(), str(), constructor()
# common data: dvr, index, length

class DviCommand:
	BOP = 0

	# DviReader, index
	def __init__(self, dvr, i):
		self.dvr = dvr
		self.index = i
		self.length = 1 # default

	def read(self, input):
		return(None)

	def __str__(self):
		return("[" + str(self.index) + "]")

	def readable(self, dvr):
		return str(self)


# -----------------------------------------------

"""
	Indices: 0 - 127

	The simplest commands are the set_char0 ...  set_char127,
	which just place a character from the current font at the current (x, y).   

	Changes h.
"""

class SetChar(DviCommand):
	def __init__(self, dvr, i):
		DviCommand.__init__(self, dvr, i)

	def __str__(self):
		return("set char " + (chr)(self.index) + "(" + str(self.index) + ")")

	# def read(self, input): # we do nothing, the character is already read
	#	return(None) 

	def getCharIndex(self):
		return(self.index)

	def execute(self):
		# place character at (h, v)
		self.dvr.execSetChar(self)

# -----------------------------------------------

"""
	Indices 128-131

	c[1,2,3,4]

	c is an Integer, the index of a character in the current (possibly very large) font.
	Otherwise acts like set_char.  TeX fonts often go to 256 characters,
		but so far I have not seen any larger than that.

"""
	
class Set(DviCommand):
	def __init__(self, dvr, i):
		DviCommand.__init__(self, dvr, i)

	def __str__(self):
		return( "set %d %d " % ( (self.index - 127), self.c) )

	def read(self, input):
		m = self.index - 127
		self.c = knuth.getUnsigned(input, m)
		self.length = m+1

	def getCharIndex(self):
		return(self.c)

	def execute(self):
		self.dvr.execSet(self)

# -----------------------------------------------

"""
	Index 132
	
	a[4] b[4]

	a, b are Lengths.  
	Sets rule.  Changes h.
"""
	
class SetRule(DviCommand):
	def __init__(self, dvr):
		DviCommand.__init__(self, dvr, SETRULE)

	def __str__(self):
		return("set rule:" + str(self.a) + " x " + str(self.b))

	def readable(self, dvr):
		return("set rule:" + Kstr(dvr.sptoadobe(a)) + " x " + Kstr(dvr.sptoadobe(b)))

	def read(self, input):
		self.a = knuth.getInt(input, 4) # height FLOAT
		self.b = knuth.getInt(input, 4) # width FLOAT
		self.length = 9

	def execute(self):
		# draw rule and shift
		self.dvr.execSetRule(self)

# -----------------------------------------------

""" 
	Indices 133-136

	c[1,2,3,4] 

	c = Integer.  Character c placed at current location, which is not changed.
"""
	
class Put(DviCommand):
	def __init__(self, dvr, i):
		DviCommand.__init__(self, dvr, i)

	def __str__(self):
		return("put " + str(self.index - 132))

	def read(self, input):
		m = self.index - 132
		self.c = knuth.getUnsigned(input, m)
		self.length = m+1

	def execute(self):
		self.dvr.execPut(self)


# -----------------------------------------------

"""
	Index 137

	a[4] b[4]

	a is width, b is height of a rule to be placed at 
	current location, which is not changed.
"""

class PutRule(DviCommand):
	def __init__(self, dvr):
		DviCommand.__init__(self, dvr, PUTRULE)

	def __str__(self):
		return("put rule:" + str(a) + " x " + str(b))

	def read(self, input):
		self.a = knuth.getInt(input, 4)
		self.b = knuth.getInt(input, 4)
		self.length = 9

	def execute(self):
		self.dvr.execPutRule(self)

# -----------------------------------------------

"""
	Index 138

	Does nothing.
"""

class Nop(DviCommand):
	def __init__(self, dvr):
		DviCommand.__init__(self, dvr,NOP)

	def __str__(self):
		return("nop")

	# def read(self, input):
	#	return(None)

	def execute(self):
		return(None)

# -----------------------------------------------

"""
	Index 139

	c[0 .. 9][4] p[4]

	Page beginning.  Set <nobr>(h, v, w, x, y, z)</nobr>
	equal to <nobr>(0, 0, 0, 0, 0, 0)</nobr>.  Set current font undefined.  
	c's are counts in TEX registers count0 etc. upon page's shipout.
	p points to previous <em>bop</em>.
"""

class Bop(DviCommand):
	# int p, int[] c
	def __init__(self, dvr):
		DviCommand.__init__(self, dvr, BOP)

	def read(self, input):
		self.c = []
		for i in range (10):
			self.c.append(knuth.getInt(input, 4))
		self.p = knuth.getInt(input, 4)
		self.length = 45

	def __str__(self):
		return("bop")

	def execute(self):
		# begins a page
		self.dvr.execBop(self)

# -----------------------------------------------

""" 
	Index 140

	End of page.  Spill out page contents.  Stack should be empty.
"""

class Eop(DviCommand):
	def __init__(self, dvr):
		DviCommand.__init__(self, dvr, EOP)

	def __str__(self):
		return("eop")

	# def read(self, input):
	#	return(None)

	def execute(self):
		self.dvr.execEop(self)

# -----------------------------------------------

"""
	Index 141

	Pushes <nobr>(h, v, w, x, y, z)</nobr> on stack, not changing them.
"""

class Push(DviCommand):
	def __init__(self, dvr):
		DviCommand.__init__(self, dvr, PUSH)

	def __str__(self):
		return("push")

	# def read(self, input):
	#	return(None)

	def execute(self):
		# stores current variables
		self.dvr.execPush(self)

# -----------------------------------------------

"""
	Index 142

	Pops <nobr>(h, v, w, x, y, z)</nobr> from stack, changing their current values.
"""

class Pop(DviCommand):
	def __init__(self, dvr):
		DviCommand.__init__(self, dvr, POP)

	def __str__(self):
		return("pop")

	# def read(self, input):
	#	return(None)

	def execute(self):
		# restores current variables
		self.dvr.execPop(self)

# -----------------------------------------------

"""
	Indices 143-146

	b[1,2,3,4] 

	Moves h right.
"""

class Right(DviCommand):
	def __init__(self, dvr, i):
		DviCommand.__init__(self, dvr, i)

	def __str__(self):
		return("right: " + str(self.b))

	def readable(self, dvr):
		return("right: " + Kstr(dvr.sptoadobe(self.b)))

	def read(self, input):
		m = self.index - 142
		self.b = knuth.getInt(input, m) # FLOAT
		self.length = 1+m

	def execute(self):
		self.dvr.execRight(self)

# -----------------------------------------------

"""
	Index 147
	
	w0 moves right current w.
	The others assign the current w as well and then move right.
"""

class W0(DviCommand):
	def __init__(self, dvr):
		DviCommand.__init__(self, dvr, W0)

	def __str__(self):
		return("w0" + ":" + str(self.dvr.w))

	# def read(self, input):
	#	return(None)

	def execute(self):
		# moves right by w
		self.dvr.execW0(self)

# -----------------------------------------------

"""
	Indices 148-151

	b[0,1,2,3,4]
	
	w0 moves right current w.
	The others assign the current w as well and then move right.
"""

class W(DviCommand):
	def __init__(self, dvr, i):
		DviCommand.__init__(self, dvr, i)

	def __str__(self):
		return("w" + str(self.index-147) + ":" + str(self.b))

	def readable(self, dvr):
		return("w" + str(self.index-147) + Kstr(dvr.sptoadobe(self.b)))

	def read(self, input):
		m = self.index - 147
		self.b = knuth.getInt(input, m) # FLOAT
		self.length = m+1

	def execute(self):
		# changes w and then moves
		self.dvr.execW(self)


# -----------------------------------------------

"""
	Index 152
	
	x0 moves right current x.
	The others assign the current x as well and then move right.
"""

class X0(DviCommand):
	def __init__(self, dvr):
		DviCommand.__init__(self, dvr, X0)

	def __str__(self):
		return("x0" + ":" + str(self.dvr.x))

	# def read(self, input):
	#	return(None)

	def execute(self):
		# moves right by x
		self.dvr.execX0(self)

# -----------------------------------------------

"""
	Indices 153-156

	b[0,1,2,3,4]

	x0 moves right current x.
	The others assign the current x as well and then move right.
"""

class X(DviCommand):
	def __init__(self, dvr, i):
		DviCommand.__init__(self, dvr, i)

	def __str__(self):
		return("x" + str(self.index - 152) + ":" + str(self.b))

	def readable(self, dvr):
		return("x" + str(self.index - 152) + ":" + Kstr(dvr.sptoadobe(self.b)))


	def read(self, input):
		m = self.index - 152
		self.b = knuth.getInt(input, m) # FLOAT
		self.length = m+1

	def execute(self):
		# changes x and then moves right by x
		self.dvr.execX(self)

# -----------------------------------------------

"""
	Indices 157-160

	a[1,2,3,4]

	Sets v = v + a - moves down (or up if negative) a units.
"""

class Down(DviCommand):
	# index = 157 - 160
	def __init__(self, dvr, index):
		DviCommand.__init__(self, dvr, index)

	def read(self, input):
		m = self.index - 156
		self.a = knuth.getInt(input, m) # FLOAT
		self.length = m+1

	def __str__(self):
		return("down" + str(self.index - 156) + ": " + str(self.a))

	def readable(self, dvr):
		return("down" + str(self.index - 156) + ": " + Kstr(dvr.sptoadobe(self.a)))

	def execute(self):
		# moves down
		self.dvr.execDown(self)

# -----------------------------------------------

"""
	Index 161

	y0 moves down current y.
	The others assign the current y as well and then move down.
"""

class Y0(DviCommand):
	def __init__(self, dvr):
		DviCommand.__init__(self, dvr, 161)

	def __str__(self):
		return("y0" + ":" + str(self.dvr.y))

	# def read(self, input):
	#	return(None)

	def execute(self):
		# moves down by y
		self.dvr.execY0(self)

# -----------------------------------------------

"""
	Indices 162-165

	a[0,1,2,3,4]

	y0 moves down current value of y.
	The others assign the current y as well and then move down.
"""

class Y(DviCommand):
	def __init__(self, dvr, i):
		DviCommand.__init__(self, dvr, i)

	def __str__(self):
		return("y" + str(self.index - 161) + ":" + str(self.a))

	def readable(self, dvr):
		return("y" + str(self.index - 161) + ":" + Kstr(dvr.sptoadobe(self.a)))

	def read(self, input):
		m = self.index - 161
		self.a = knuth.getInt(input, m) # FLOAT
		self.length = m+1

	def execute(self):
		# changes y and then moves down by y
		self.dvr.execY(self)

# -----------------------------------------------

"""
	Index 166

	z0 moves down current z.
	The others assign the current z as well and then move down.
"""

class Z0(DviCommand):
	def __init__(self, dvr):
		DviCommand.__init__(self, dvr, Z0)

	def __str__(self):
		return("z0" + ":" + str(self.dvr.z))

	# def read(self, input):
	#	return(None)

	def execute(self):
		# moves down by z
		self.dvr.execZ0(self)

# -----------------------------------------------

"""
	Indices 167-170

	a[0,1,2,3,4]

	z0 moves down current z.
	The others assign the current z as well and then move down.
"""

class Z(DviCommand):
	def __init__(self, dvr, i):
		DviCommand.__init__(self, dvr, i)

	def __str__(self):
		return("z" + str(self.index - 166) + ":" + str(self.a))

	def readable(self, dvr):
		return("z" + str(self.index - 166) + ":" + Kstr(dvr.sptoadobe(self.a)))

	def read(self, input):
		m = self.index - 166
		self.a = knuth.getInt(input, m) # FLOAT
		self.length = m+1

	def execute(self):
		# changes z and then moves down by z
		self.dvr.execZ(self)

# -----------------------------------------------

""" 
	Indices 171-234

	Sets current font to font index - 171.
"""

class FntNum(DviCommand):
	def __init__(self, dvr, i):
		DviCommand.__init__(self, dvr, i)

	def __str__(self):
		return("font num " + str(self.index-171))

	# def read(self, input):
	#	return(None)

	def execute(self):
		self.dvr.execFntNum(self)

# ---------------------------------------------

"""
	Indices 241-242

	k[1,2,3,4]

	Sets current font equal to font k (used for k over 63) 
	(extension of FntNum, really).
"""

class Fnt(DviCommand):
	def __init__(self, dvr, i):
		DviCommand.__init__(self, dvr, i)

	def __str__(self):
		return("fnt " + str(self.k))

	def read(self, input):
		m = self.index - 234
		self.k = knuth.getUnsigned(input, m)
		# print "k =", self.k
		self.length = m+1

	def execute(self):
		self.dvr.execFnt(self)

# -----------------------------------------------

"""
	Indices 239-242

	k[1,2,3,4] x[k]

	This is for specials.  The array x is by default passed unread as a string
"""

class XXX(DviCommand):
	def __init__(self, dvr, i):
		DviCommand.__init__(self, dvr, i)

	def __str__(self):
		return("xxx ")

	def read(self, input):
		m = self.index - 238
		k = knuth.getUnsigned(input, m)
		self.x = knuth.getString(input, k)
		self.length = 1+m+k

	def execute(self):
		self.dvr.doSpecial(self.x)

# -----------------------------------------------

"""
	Indices 243-246

	k[1,2,3,4] c[4] s[4] d[4] a[1] ell[1] n[a+ell]

	k is the font number, 
	a length of directory, ell length of name - both stored in array n
	c is a checksum, s is scale factor in sp units, d is design size.
	Thus font k is to be used at (mag/1000)*(s/d) times normal size. 
"""

import os

class FntDef(DviCommand):
	def __init__(self, dvr, i):
		DviCommand.__init__(self, dvr, i)

	def __str__(self):
		if (len(self.directory) > 0):
			s = "font def " + str(self.k) + ": " + os.path.join(self.directory, self.fontfile)
		else:
			s = "font def " + str(self.k) + ": " + self.fontfile
		s += " " + str(self.s) + ", " + str(self.d)
		return(s)

	def readable(self, dvr):
		if (len(self.directory) > 0):
			s = "font def " + str(self.k) + ": " + os.path.join(self.directory, self.fontfile)
		else:
			s = "font def " + str(self.k) + ": " + self.fontfile
		s += " " + Kstr(dvr.sptoadobe(self.s)) + ", " + Kstr(dvr.sptoadobe(self.d))
		return(s)


	def read(self, input):
		m = self.index - 242
		# print "m = ", m
		self.k = knuth.getUnsigned(input, m)
		self.c = knuth.getUnsigned(input, 4)
		self.s = knuth.getUnsigned(input, 4) # FLOAT
		self.d = knuth.getUnsigned(input, 4) # FLOAT
		a = (ord)(input.read(1))
		ell = (ord)(input.read(1))
		self.directory = knuth.getString(input, a)
		self.fontfile = knuth.getString(input, ell)
		self.length = m+15+a+ell

	def execute(self):
		self.dvr.execFntDef(self)

# -----------------------------------------------

"""
	Index 247

	i[1] num[4] den[4] mag[4] k[1] x[k]
	
	Starts file.  i is id (always 2),
	k is length of comment x

	The fraction num/den specifies the size of a unit in meters.  We multiply 
	dimensions by this to get sizes in multiples of 10^{-7} meter.
	if the unit is a Knuth sp = pt*2^{-16} = (2.54/72.27) * 2^{-16} cm ...
	num = 25400000, den = 473628672, fraction = (2.54 10^{7}) / (72.27 2^{16}).
"""
	
class PreAmble(DviCommand):
	def __init__(self, dvr):
		DviCommand.__init__(self, dvr, PRE)

	def __str__(self):
		if (len(self.x) > 0):
			s = "preamble: " + self.x
		else:
			s = "preamble"
		# s += ":" + str(self.num) + ", " + str(self.den)
		return(s)

	def read(self, input):
		self.i = input.read(1)
		self.num = knuth.getUnsigned(input, 4) # FLOAT
		self.den = knuth.getUnsigned(input, 4) # FLOAT
		self.mag = knuth.getUnsigned(input, 4) # FLOAT
		k = (ord)(input.read(1))
		self.x = knuth.getString(input, k)
		self.length = 15+k

	def execute(self):
		# begins the file
		# read num, den, magnification
		self.dvr.execPreAmble(self)

# -----------------------------------------------

"""
	Index 248

	p[4] num[4] den[4] mag[4] ell[4] u[4] s[2] t[2]

	Postamble, at end of file.
	p is pointer to final bop,
	num, den, mag duplicate data in the preamble,
	ell and u give height + depth of tallest page, 
	width of widest page (usually ignored),
	s is maximum stack depth of pushes & pops,
	t is total number of bops (i.e. pages).
"""

class PostAmble(DviCommand):
	def __init__(self, dvr):
		DviCommand.__init__(self, dvr, POST)

	def __str__(self):
		return("postamble")

	def read(self, input):
		self.p = knuth.getInt(input, 4)
		self.num = knuth.getInt(input, 4) # FLOAT
		self.den = knuth.getInt(input, 4) # FLOAT
		self.mag = knuth.getInt(input, 4) # FLOAT
		self.ell = knuth.getInt(input, 4) # FLOAT
		self.u = knuth.getInt(input, 4) # FLOAT
		self.s = knuth.getUnsigned(input, 2)
		self.t = knuth.getUnsigned(input, 2)
		self.length = 29

	def execute(self):
		self.dvr.execPostAmble(self)

# -----------------------------------------------

"""
	Index 249

	q[4] i[1] 223's (four or more, for padding)

	q is pointer to the post command that started the postamble,
	i is id (the type of dvi file, currently always 2).
"""
	
class PostPostAmble(DviCommand):
	def __init__(self, dvr):
		DviCommand.__init__(self, dvr, POSTPOST)

	def __str__(self):
		return("postpostamble")

	def read(self, input):
		self.q = knuth.getInt(input, 4)
		self.i = knuth.getInt(input, 1)
		c = (ord)(input.read(1))
		f = 0
		while (((c + 256) & 255) == 223):
			f += 1
			c = input.read(1)
			if c:
				c = (ord)(c)
			else:
				break
		# if not the right number of 223's?
		self.length = 6+f

	def execute(self):
		self.dvr.execPostPostAmble(self)


# =====================================================================================
# =====================================================================================
# =====================================================================================


"""
	Multiple indices

	  Placeholder for illegal commands in a virtual font
	"""
	
class VIllegal(DviCommand):
	def __init__(self, dvr,k):
		DviCommand.__init__(self, dvr, k)

	def __str__(self):
		return("Illegal VFont command %d" % self.index)

	def read(self, input):
		logging.error(str(self)) 

	def execute(self):
		logging.error(str(self)) 


# =========================================================================
# === DVIREADER ===========================================================
# =========================================================================


"""
	DviReader reads commands from the .dvi file
	and execute them. The basic DviReader just
	keeps track of the current state,
	without doing anything else.
	Extensions will display the file in various media.

	The most complicated structures it deals with are DviFonts.
"""

class DviReader:

	"""
	The current state:

		* int h, v, w, x, y, z 
	
		These are positions, measured in the units 
		specified by num, den relative to sp.
		In particular, h and v are the current 
		horizontal and vertical coordinates.

		* Stack dims
		
		Stores dimension arrays.  Used as a primnitive graphisc state
		in PostScript.  Here implemented as a list.  It is started all over
		at the beginning of each page.

		* DviFont currentFont
		* int currentPageNo

		In some versions, we need to access pages independently.

		* int num, den, mag
		Variables set once at the start of every file. 

		* Stack fonts

		Used to assemble a list of all fonts.

		* Stack currentPage

		Used to assemble the current page. */

		* Hashtable fontTable

		Records fonts so far defined.  Implemented as a dictionary.

		A list of the fonts that haven't been located. */

		* InputStream input
	
		* DviCommand[] command
"""
	
	def __init__(self, file):
		# new version: read entire file, then process that
		# load byte array
		# self.input = open(file + ".dvi", "rb")
		# self.dvi = self.input.read()
		# self.input.close()
		
		self.input = open(file + ".dvi", "rb")
		self.mkCommands()
		self.fontTable = {}
		currentPageNo = -1
		self.descendingVirtualFonts = True
		self.scaleFactor = 1
		self.vFontStates = []
		self.currentFont = None
		self.h = 0
		self.v = 0
		self.w = 0
		self.x = 0
		self.y = 0
		self.z = 0
		self.minwdSet = False
		self.maxwdSet = False
		self.maxhtSet = False
		self.minhtSet = False
		self.maxlevelSet = False
		P = []
		c = self.getpre()
		self.mag = c.mag
		c = self.getpostpost()
		n = c.q
		P.append(n)
		c = self.getcommandat(c.q)
		n = c.p
		while not n == -1:
			P.append(n)
			c = self.getcommandat(n)
			n = c.p
		self.page = []
		for i in range(len(P)):
			self.page.append(P[i-1])
		self.input.close()
		self.input = open(file + ".dvi", "rb")
		
	def getpre(self):
		self.input.seek(0,0)
		c = self.getCommand()
		return(c)
	
	def getpostpost(self):
		k = 1
		while(True):
			self.input.seek(-k,2)
			s = ord(self.input.read(1))
			if not s == 223:
				break
			k += 1
		self.input.seek(-k-5,2)
		c = self.getCommand()
		return c
		
	def getcommandat(self,n):
		self.input.seek(n, 0)
		c = self.getCommand()
		return c
	
	
	def readpage(self, n):
		pass

	def mkCommands(self):
		command = [] # 0 --- 249
		for i in range(128):
			command.append(SetChar(self, i))

		for i in range(128, 132):
			command.append(Set(self, i))

		command.append(SetRule(self))
		for i in range(133, 137):
			command.append(Put(self, i))

		command.append(PutRule(self))
		command.append(Nop(self))
		command.append(Bop(self))
		command.append(Eop(self))
		command.append(Push(self))
		command.append(Pop(self))
		for i in range(143, 147):
			command.append(Right(self, i))

		command.append(W0(self))
		for i in range(148, 152):
			command.append(W(self, i))

		command.append(X0(self))
		for i in range(153, 157):
			command.append(X(self, i))

		for i in range(157, 161): # OK
			command.append(Down(self, i))

		command.append(Y0(self)) # 161
		for i in range(162, 166):
			command.append(Y(self, i))

		command.append(Z0(self))
		for i in range(167, 171):
			command.append(Z(self, i))

		for i in range(171, 235):
			command.append(FntNum(self, i))

		for i in range(235, 239):
			command.append(Fnt(self, i))

		for i in range(239, 243):
			command.append(XXX(self, i))

		for i in range(243, 247):
			command.append(FntDef(self, i))

		command.append(PreAmble(self))
		command.append(PostAmble(self))
		command.append(PostPostAmble(self))

		# Create a copy of the command list for use when processing
		# virtual font dvi streams
		vfont_command = [ c for c in command ]
		vfont_command[139] = VIllegal(self,139)
		vfont_command[140] = VIllegal(self,140)
		for i in range(243,248):
			vfont_command[i]=VIllegal(self,i)
		
		self.standard_command = command
		self.vfont_command = command
		self.command = self.standard_command

	def render(self):
		while (True):
			c = self.getCommand()
			c.execute()
			P = self.getCurrentPosition()
			# print "["  + str(c.index) + "]:" + c.readable(self)
			print c.readable(self)
			print "\t(" + Kstr(self.sptoadobe(P[0])) + ", " + Kstr(self.sptoadobe(P[1])) + ")"
			if (c.index == self.EOF()):
				break

	def getCommand(self):
		while( (len(self.vFontStates) > 0) and (self.input.pos>=self.input.len) ):
			self.popVFont()
		n = ord(self.input.read(1))
		# print "n = ", n
		self.command[n].read(self.input)
		return(self.command[n])

	def getCurrentPosition(self):
		return([self.h, self.v])

	def EOF(self):
		return(POST)

	def close(self):
		self.input.close()

	def sptoadobe(self, x):
		return(x*(72/72.27)*self.mag/(1000.0*(1 << 16)))

	def Str(x):
		return("%3.4f" % x)
	Str=staticmethod(Str)

	# adjust minimum width etc.
	def adjustbbox(self, c): # NOTE: only character index occurs here
		# done just *after* the character is set
		loc = self.getCurrentPosition()
		w = self.currentFont.getCharWidth(c)
		h = self.currentFont.getCharHeight(c)
		d = self.currentFont.getCharDepth(c)
		ss = self.currentFont.scaledSize
		if not self.maxlevelSet:
			self.maxlevel = loc[1]
			self.maxlevelSet = True
		else:
			level = loc[1]
			if (level > self.maxlevel):
				self.maxlevel = level
			# print "min level now = " + str(self.pageHeight - self.sptoadobe(level))
		# print "wd, char width, font's scaled size", self.sptoadobe(loc[0]), w, self.sptoadobe(ss)
		if not self.minwdSet:
			self.minwd = loc[0] - w*ss
			self.minwdSet = True
		else:
			wd = loc[0] - w*ss
			if (wd < self.minwd):
				self.minwd = wd
		if not self.maxwdSet:
			self.maxwd = loc[0]
			self.maxwdSet = True
		else:
			wd = loc[0]
			if (wd > self.maxwd):
				self.maxwd = wd
		if not self.minhtSet:
			self.minht = loc[1]-h*ss;
			self.minhtSet = True
		else:
			ht = loc[1]-h*ss
			if (ht < self.minht):
				self.minht = ht
		if not self.maxhtSet:
			self.maxht = loc[1]+d*ss
			self.maxhtSet = True
		else:
			ht = loc[1]+ss*d
			if (ht > self.maxht):
				self.maxht = ht
				
	def startFont( self, font ):
		self.currentFont = font
		if( not self.currentFont.isUsed ):
			self.currentFont.load()

	def pushVFont( self, cindex):
		vFont = self.currentFont.vChars
		self.dims.append([ self.h, self.v, self.w, self.x, self.y, self.z])
		self.vFontStates.append([ self.dims, self.currentFont, self.scaleFactor, self.fontTable, self.input ])
		self.w=0; self.x=0; self.y=0; self.z=0;
		self.fontTable = vFont.fontTable
		self.scaleFactor *= (self.currentFont.scaledSize*1.0/(1 << 20))
		self.currentFont = vFont.defaultFont
		if( self.currentFont != None ):
			self.startFont( self.currentFont );
		self.dims=[]
		self.input = StringIO.StringIO(vFont.packets[cindex].dvi)
		self.command = self.vfont_command

	def popVFont( self ):
		self.input.close()
		[ self.dims, self.currentFont, self.scaleFactor, self.fontTable, self.input ] = self.vFontStates.pop()
		[ self.h, self.v, self.w, self.x, self.y, self.z ] = self.dims.pop()
		self.command = self.standard_command

	# --- the various execution methods ---------------------------

	def execSetChar(self, dvc):
		# place character n at (h, v)

		cw = self.currentFont.getCharWidth(dvc.index)
		ss = self.currentFont.scaledSize
		dh = cw*ss*self.scaleFactor
		# print "dh = " + str(dh)
		self.h += dh
		if( self.currentFont.isVirtual and self.descendingVirtualFonts):
			self.pushVFont( dvc.index )
			# Before we saved the current state, we translated as if we had 
			# rendered the character.  We undo this here to start drawing the 
			# current character.
			self.h-=dh


	def execFntNum(self, dvc):
		# print "setting font " + str(dvc.index-171)
		self.startFont( self.fontTable.get("f" + str(dvc.index-171) ) )

	def execFnt(self, dvc):
		# changes current font
		self.startFont( self.fontTable.get("f" + str(dvc.k)) )

	def execSetRule(self, dvc):
		# draw rule and shift
		self.h += dvc.b*self.scaleFactor

	def execPutRule(self, dvc):
		# draw rule but don't shift
		return(None)

	def execSet(self, dvc):
		cw = self.currentFont.getCharWidth(dvc.c)
		ss = self.currentFont.scaledSize
		dh = cw*ss*self.scaleFactor
		self.h += dh
		if( self.currentFont.isVirtual and self.descendingVirtualFonts):
			self.pushVFont( dvc.c )
			# Before we saved the current state, we translated as if we had 
			# rendered the character.  We undo this here to start drawing the 
			# current character.
			self.h-=dh

	def execPut(self, dvc):
		return(None)

	def execBop(self, dvc):
		# begins a page
		self.h = 0
		self.v = 0
		self.w = 0
		self.x = 0
		self.y = 0
		self.z = 0
		self.dims = []
		self.currentFont = None
		# read the current page number from some c[i] - see TEX book
		self.currentPage = []

	def execFntDef(self, dvc):
		key = "f" + str(dvc.k)
		# System.err.println("key = " + key)
		if (self.fontTable.get(key) == None):
			tf = DviFont(dvc.fontfile, dvc.s, dvc.d, dvc.k)
			self.fontTable[key] = tf

	def execEop(self, dvc):
		return(None)

	def execPush(self, dvc):
		# stores current variables
		V = [ self.h, self.v, self.w, self.x, self.y, self.z ]
		self.dims.append(V)

	def execPop(self, dvc):
		# restores current variables
		self.h, self.v, self.w, self.x, self.y, self.z = self.dims.pop()

	def execPreAmble(self, dvc):
		# begins the file
		# read num, den, magnification
		self.num = dvc.num
		self.den = dvc.den
		self.mag = dvc.mag
		self.ell = dvc.mag/1000.0

	def execXXX(self, dvc):
		# holds a single byte array x
		s = dvc.getString()
		self.doSpecial(s)

	def execRight(self, dvc):
		# moves right
		self.h += dvc.b*self.scaleFactor

	def execDown(self, dvc):
		# moves down
		self.v += dvc.a*self.scaleFactor

	def execW0(self, dvc):
		# moves right by w
		self.h += self.w*self.scaleFactor

	def execW(self, dvc):
		# changes w and then moves
		self.w = dvc.b*self.scaleFactor
		self.h += self.w

	def execX0(self, dvc):
		# moves right by x
		self.h += self.x

	def execX(self, dvc):
		# changes x and then moves right by x
		self.x = dvc.b*self.scaleFactor
		self.h += self.x

	def execY0(self, dvc):
		# moves down by y
		self.v += self.y

	def execY(self, dvc):
		# changes y and then moves down by y
		self.y = dvc.a*self.scaleFactor
		self.v += self.y

	def execZ0(self, dvc):
		# moves down by z
		self.v += self.z

	def execZ(self, dvc):
		# changes z and then moves down by z
		self.z = dvc.a*self.scaleFactor
		self.v += self.z

	def execPostAmble(self, dvc):
		return(None)

	def execPostPostAmble(self, dvc):
		return(None)

	# --------------------------------------------------------------

	def doSpecial(self, s):
		# all of my admissible specials will be command{}{} ...
		# print "------------------------ special: " + s + " -------------------------"

		""" old java code:

		stringInputStream si = new stringInputStream(s)
		String type = si.getWord()
		System.err.println("type = " + type)
		if ("applet".equals(type)):
			Hashtable h = new Hashtable()
			String w = si.getWord()
			int count = 0
			while (!"".equals(w) && count++ < 8):
				String p = stringInputStream.getPrefix(w, '=')
				String q = stringInputStream.getSuffix(w, '=')
				h.put(p, q)
				w = si.getWord()
				System.err.println("w = " + w)
		
			System.err.println("width = " + (String) h.get("width"))
			System.err.println("height = " + (String) h.get("height"))
			System.err.println("code = " + (String) h.get("code")) """
	
	
	

# ========================================================================
